(*
  The BDRestoration package, version 1.0.
*)

(*
  This package implements a ``blind'' restoration
  algorithm; it was created by Luca Caucci. You
  may send any comment, suggestion or idea to
  caucci@cs.unibo.it.
*)

(*
  HISTORY:
    2002-07-18: Version 1.0 released
    2002-09-26: Routine showSupport added
    2002-12-04: Routine computeSNR added
    2003-02-01: Option NSHRIFRestoredList added
    2003-02-03: Routine evaluateSupport added
*)

BeginPackage["BDRestoration`"]

showImage::usage = "showImage[m] shows a picture
corresponding to the specified matrix m. The real
part of each element of m is taken, scaled in the
range [0, 1] and then shown as a raster picture.
The function returns the Graphics object drawn.";

showSupport::usage = "showSupport[m, ny, nx] shows a
picture and its corresponding support. The picture
is provied to this function via the matrix m, while
the vertical and horizontal support dimensions are
specified via the parameters ny and nx, respectively.
The support is depicted as a box with white dashed
borders. The function returns the Graphics object
drawn.";

extendImageMatrix::usage = "extendImageMatrix[m,
my, mx] extends both the rows and the columns of
the given matrix m adding mx zeros to both sides of
each row and my zeros to both sides of each column.";

extendPSFMatrix::usage = "extendPSFMatrix[m, ny,
nx] extends the PSF matrix m by extending each row
and each column and rearranging its elements. The
dimensions of the matrix must be odd and the number
of rows and columns will grow of my - 1 and mx - 1,
respectively. The matrix is extended is such a
way that it can be used for convolutions.";

lowPassMatrixFilte::usage = "lowPassMatrixFilter[m,
fCutOff, deltaU, deltaV] implements a lowpass filter.
It can be used to cut those components beyond the
cutoff frequency fCutOff of the observation system;
in this way some noise can be removed from the observed
image matrix m. deltaU and deltaV are the frequency
steps along each axis of the Fourier spectrum of the
image.";

getFilterEstimation::usage = "getFilterEstimation[m,
threshold] can be used to obtain a good estimation of
the inverse filter that is then used to initialize the
inverse filter matrix of the restoration algorithm. m
is the noisy guide star image and threshold tells whose
spectral components must be ignored in order to avoid
noise amplication when using the estimated inverse
filter.";

convolveMatrices::usage = "convolveMatrices[f, h]
convolves the matrices f and h after having extended
(i.e. padded with zeros) them. This function can be
used to create blurred test images to restore.";

computeSNR::usage = "computeSNR[s, n] computes the
signal-to-noise ratio between two matrices. The matrix
s represents the signal while n the noise affecting it.
The SNR value returned is expressed in Decibel (dB).";

evaluateSupport::usage = "evaluateSupport[g, perc] is
used to evaluate the support of the image to restore.
The image is expressed by means of the matrix g while
perc is a percent threshold that tells the maximum
intensity level of the points that should be considered
background. This function returns a list of two
integers which represent the estimated image support.";

NSHRIF::usage = "NSHRIF[gTilde, sStar, param, nyTilde
nxTilde] implements the NSH-RIF restoration algorithm.
gTilde is the observed noisy image, sStar is the
estimated inverse filter, param is the minimization
parameter vector while nyTilde and nxTilde define
the estimated support. If the input is valid, the
function returns a list l of three or four elements: the
first element of l is the list of the images restored
at each iteration if the option NSHRIFRestoredList
is set to True otherwise is the image restored at the
last iteration, the second element of l contains the
values of the cost function at each iteration while
the third element of l contains the relative variation
of the cost function trough the iterations. If the
optional parameter NSHRIFTrueF is not set to the empty
list, l will contain a fourth parameter, that can be
used to test the effectiveness of the algorithm: if
NSHRIFTrueF is set to the true, non-blurred image,
then the relative error between the restored and the
true image is evaluated at each iteration and the
list of such values is returned as the fourth element
of the list l.";

NSHRIFMaxIterations::usage = "NSHRIFMaxIterations is
an option for NSH-RIF and allows to set the upper bound
on the number of iteration. Its default value is 100.";

NSHRIFEpsilon::usage = "NSHRIFEpsilon is an option for
NSH-RIF and represents a tollerance for the minimization
algorithm. If the value of the cost function drops below
NSHRIFEpsilon the iterative restoration algorithm stops.
The default value is 10^-6.";

NSHRIFRestoredList::usage = "NSHRIFRestredList is an
option for NSH-RIF and indicates whether the image
restored at each iteration should be stored or not.
If this option is set to True, the image restored at
each iteration is appended to a list and this list is
returned by the NSHRIF function. If NSHRIFRestredList
is set to False, which is the default value, only the
last image is returned as the result of the restoration
process.";

NSHRIFTrueF::usage = "NSHRIFTrueF is an option for NSHRIF
and can be used to measure the effectiveness of the
restoration algorithm by running it on some test images.
NSHRIFTrueF lets us tell the restoration algorithm
the true non-blurred image: in this way, a relative
error between the true and the restored image can be
computed for testing purposes. Its default value is the
empty list which means we can't/don't want to compute
such an error.";

NSHRIFLowPassFilter::usage = "NSHRIFLowPassFilter is an
option for NSHRIF and if it is set to the list {fCutOff,
deltaU, deltaV}, then a low-pass filter is applied to
the observed image passed to NSHRIF. fCutOff is the
cutoff frequency of the observation system; deltaU
and deltaV are the frequency steps for the pixels
along each axis. The default value is the empty list
which means that the low-pass filter is not applied
to the observed image. Low-passing the image can
sometimes help remove some noise.";

NSHRIFStopEpsilon::usage = "NSHRIFStopEpsilon is an
option for NSHRIF and represents a tollerance for the
relative variation of the value of the cost function
through the iterations of the minimization algorithm. If
such a relative variation drops below NSHRIFStopEpsilon,
the algorithm returns. 10^-6 is the default value for
NSHRIFStopEpsilon.";

totalVariation::usage = "totalVariation[y, nu]
implements the total variation (TV) regularization
of the image. It can be used to sharpen a restored
image. Its inputs are the image y to regularize and
the TV-regularization parameter vector nu whose
length should be two. The option TotalVariationTrueF
allows to log the relative errors between the true
image and the TV-regularized one obtained at each
iteration of the total variation regularization
algorithm. totalVariation returns the regularized image
and the list of relative errors if TotalVariationTrueF
is not set to the empty list.";

TotalVariationTrueF::usage = "TotalVariationTrueF
is an option for totalVariation used to test the
effectiveness of the total variation regularization
algorithm. If it is not set to the empty list, the
relative error between the restored image and the
true image (i.e. the matrix associated to this
option) is computed at each iteration of the total
variation regularization algorithm and the list of
all such errors is returned by the total variation
regularization routine. The default value for this
option is the empty list.";

TotalVariationEpsilon::usage = "TotalVariationEpsilon
is an option for totalVariation and specifies the
tollerance used by the total variation regularization
routine. Its default value is 10^-8.";

Begin["`Private`"]

(* error messages *)
BDRestoration::EvenVectorLength = "The length of the vector must be odd.";
BDRestoration::EvenMatrixDimensions = "Both dimensions of the matrix must be odd.";
BDRestoration::InvalidLength = "The length specified is not valid.";
BDRestoration::InvalidDimensions = "The dimensions specified are not valid.";
BDRestoration::InvalidSupport = "The support of the image is not valid.";
BDRestoration::InvalidParameters = "`1` are not valid parameters.";
BDRestoration::InvalidImage = "Invalid observed image.";
BDRestoration::InvalidParamNumber = "Invalid number of restoration/regularization parameters.";
BDRestoration::InvalidInverseFilterDimensions = "Invalid inverse filter dimensions.";
BDRestoration::InverseFilterTooBig = "Inverse filter matrix too big.";
BDRestoration::SupportTooBig = "Image support too big.";

(*
  This function shows a pictured corresponding
  to the specified matrix. The real part of each
  element of the matrix is taken, scaled in the
  range [0, 1] and then shown as a raster picture.
  The function returns the Graphics object drawn.
*)
showImage[m_?MatrixQ] :=
  Module[{newMatr, min, max},
    newMatr = N[Re[m]];
    min = Min[newMatr];
    max = Max[newMatr];
    Show[Graphics[Raster[(newMatr - min) / (max - min)]], AspectRatio -> Automatic, ImageSize -> 2 * 72]
  ]

(*
  This function shows a picture and its corresponding
  support. The picture is provied to this function via
  the matrix m, while the vertical and horizontal support
  dimensions are specified via the parameters ny and nx,
  respectively. The support is depicted as a box with
  white dashed borders. The function returns the Graphics
  object drawn.
*)
showSupport[m_?MatrixQ, ny_Integer, nx_Integer] :=
  Module[{ly, lx, min, max, realM, lxm, lxp, lym, lyp},
    {ly, lx} = Dimensions[m];
    realM = N[Re[m]];
    min = Min[realM];
    max = Max[realM];
    lxm = (lx - nx) / 2;
    lxp = (lx + nx) / 2;
    lym = (ly - ny) / 2;
    lyp = (ly + ny) / 2;
    If[IntegerQ[lym] && IntegerQ[lxm],
      Show[Graphics[{Raster[(realM - min) / (max - min)], Dashing[{0.05, 0.03}], RGBColor[1, 1, 1], Line[{{lxm, lym},
        {lxp, lym}, {lxp, lyp}, {lxm, lyp}, {lxm, lym}}]}], AspectRatio -> Automatic, ImageSize -> 2 * 72]
    ]
  ]

(*
  This function pads a vector adding ``m'' zeros
  before its elements and other ``m'' zeros after
  them.
*)
extendImageVector =
  Compile[{{v, _Real, 1}, {m, _Integer}},
    Module[{lenV},
      If[m >= 0,
        lenV = Length[v];
        PadLeft[PadRight[v, m + lenV], 2 * m + lenV],
        (* invalid length error *)
        Message[BDRestoration::InvalidLength];
        {}
      ]
    ]
  ];

(*
  This function extends a PSF vector rearranging
  its elements and adding ``n - 1'' zeros. The
  length of the PSF vector should be odd. The vector
  is extended is such a way that it can be used for
  convolutions.
*)
extendPSFVector =
  Compile[{{psf, _Real, 1}, {n, _Integer}},
    Module[{lenPSF},
      lenPSF = Length[psf];
      If[OddQ[lenPSF],
        If[n >= 1,
          RotateLeft[PadRight[psf, lenPSF - 1 + n], (lenPSF - 1) / 2],
          (* invalid length error *)
          Message[BDRestoration::InvalidLength];
          {}
        ],
        (* an error message is issued *)
        Message[BDRestoration::EvenVectorLength];
        (* if the length of the vector is even, 
        the empty list is returned *)
        {}
      ]
    ]
  ];

(*
  This function extends both the rows and the columns
  of the given matrix ``m'' adding ``my'' zeros to
  both sides of each column and ``mx'' zeros to both
  sides of each row.
*)
extendImageMatrix =
  Compile[{{m, _Real, 2}, {my, _Integer}, {mx, _Integer}},
    Module[{i, j, ny, nx},
      If[(my >= 0) && (mx >= 0),
        {ny, nx} = Dimensions[m];
        (* extend along both directions *)
        If[my != 0,
          Join[
            Table[0.00, {i, my}, {j, 2 * mx + nx}],
            Table[extendImageVector[m[[i]], mx], {i, ny}],
            Table[0.00, {i, my}, {j, 2 * mx + nx}]
          ],
          Table[extendImageVector[m[[i]], mx], {i, ny}]
        ],
        (* invalid dimensions error *)
        Message[BDRestoration::InvalidDimensions];
        {{}}
      ]
    ]
  ];

(*
  This function extends a PSF matrix extending each
  row and each column and rearranging its elements.
  The dimensions of the matrix must be odd and the
  number of rows and columns grows of ``my - 1'' and
  ``mx - 1'', respectively. The matrix is extended is
  such a way that it can be used for convolutions.
*)
extendPSFMatrix =
  Compile[{{psf, _Real, 2}, {ny, _Integer}, {nx, _Integer}}, 
    Module[{i, j, my, mx, py, px},
      (* the dimensions of the matrix must be odd *)
      {py, px} = Dimensions[psf];
      If[OddQ[py] && OddQ[px],
        If[(ny >= 1) && (nx >= 1),
          {my, mx} = ({py, px} - 1) / 2;
          (* extend along both directions *)
          If[ny != 1,
            RotateLeft[Join[Table[extendPSFVector[psf[[i]], nx], {i, 2 * my + 1}],
              Table[0.00, {i, ny - 1}, {j, 2 * mx + nx}]], my],
            RotateLeft[Join[Table[extendPSFVector[psf[[i]], nx], {i, 2 * my + 1}]], my]
          ],
          (* invalid dimensions error *)
          Message[BDRestoration::InvalidDimensions];
          {{}}
        ],
        (* an error message is emitted *)
        Message[BDRestoration::EvenMatrixDimensions];
        (* if one or both dimensions of the matrix are
        even, the empty matrix is returned *)
        {{}}
      ]
    ]
  ];

(*
  This function implements a lowpass filter. It can
  be used to cut those components beyond the cutoff
  frequency of the observation system; in this way
  some noise can be removed from the observed image
  matrix ``m''.
*)
lowPassMatrixFilter =
  Compile[{{m, _Real, 2}, {fCutOff, _Real}, {deltaU, _Real}, {deltaV, _Real}},
    Module[{i, j, numRows, numCols, ry, rx, result},
      {numRows, numCols} = Dimensions[m];
      {ry, rx} = Floor[{numRows, numCols} / 2];
      (* compute and shift the spectrum of the image *)
      result = Fourier[Table[If[EvenQ[i + j], m[[i, j]], -m[[i, j]]], {i, numRows}, {j, numCols}]];
      (* set to zero all the components beyond the cutoff frequencdy *)
      Do[
        (* NOTE : You'd better not remove ``I * 0.00''
        because with it we can avoid a run-time type
        conversion and so save a lot of CPU time! *)
        If[(((i - ry) * Abs[deltaV])^2  + ((j - rx) * Abs[deltaU])^2) >= fCutOff^2,
          result[[i, j]] = 0.00 + I * 0.00;
        ],
      (* that is done for each element of the spectrum matrix *)
      {i, numRows}, {j, numCols}];
      (* compute the image from the truncated spectrum *)
      result = Re[InverseFourier[result]];
      (* shift it back *)
      Re[Table[If[EvenQ[i + j], result[[i, j]], -result[[i, j]]], {i, numRows}, {j, numCols}]]
    ], {{Fourier[_], _Complex, 2}, {InverseFourier[_], _Complex, 2}}
  ];

(*
  This function truncates the spectrum components
  of a block circulant matrix whose magnitudes are
  bigger than ``alpha''. It is used to penalize those
  Fourier components whose moduli are too small.
*)
truncateComponents =
  Compile[{{m, _Complex, 2}, {alpha, _Real}},
    Map[Function[x, If[Abs[x] <=  alpha, x, 0.00]], m, {2}]
  ];

(*
  This function can be used to obtain a good estimation
  of the inverse filter that is then used to initialize
  the inverse filter matrix of the restoration algorithm.
*)
getFilterEstimation =
  Compile[{{guideStar, _Real, 2}, {zetaTrunc, _Real}},
    Module[{eigen, mul, ty, tx, myHat, mxHat, i, j, inverseFilter},
      {ty, tx} = Dimensions[guideStar];
      (* we know that ``myHat'' and ``mxHat'' are integer
      but we need to force Mathematica to know that *)
      {myHat, mxHat} = Floor[({ty, tx} - 1) / 2];
      (* we save a bit of computation! *)
      mul = Sqrt[ty * tx];
      (* compute the eigenvalues using the Fourier transform *)
      eigen = Fourier[guideStar] *  mul;
      (* replace those who are too small with an ``identity''
      component. This can help avoid noise amplification *)
      eigen = Table[If[Abs[eigen[[i, j]]] <= zetaTrunc, mul * Exp[-(I * 2 * Pi) * (myHat * (i - 1) /
        ty + mxHat * (j - 1) / tx)], eigen[[i, j]]], {i, 2 * myHat + 1}, {j, 2 * mxHat + 1}];
      (* compute the inverse filter from the truncated eigenvalues *)
      inverseFilter = Re[InverseFourier[1.00 / eigen]] / mul;
      (* force the energy conservation constraint
      on the estimated inverse matrix filter *)
      inverseFilter / Apply[Plus, inverseFilter, {0, 1}]
    ], {{Fourier[_], _Complex, 2}, {InverseFourier[_], _Complex, 2}}
  ];

(*
  This function imposes the non-negativity and support
  constraints on the estimated image ``y''. The estimated
  support must be provided as well and is passed to the
  function through the formal parameters ``nyTilde'' and
  ``nxTilde''.
*)
supportConstraints =
  Compile[{{y, _Real, 2}, {nyTilde, _Integer}, {nxTilde, _Integer}},
    Module[{ly, lx, myTilde, mxTilde, i, j},
      {ly, lx} = Dimensions[y];
      {myTilde, mxTilde} = ({ly, lx} - {nyTilde, nxTilde}) / 2;
      (* check whether the support is valid *)
      If[(FractionalPart[myTilde] == 0) && (FractionalPart[mxTilde] == 0) && (Min[myTilde, mxTilde] >= 0),
        (* a matrix whose nonzero elements corresponding
        to the pixels of the estimated image that do not
        violate the constraints is computed and returned *)
        (* replace the ``bad'' pixels with zeros *)
        Table[If[(myTilde + 1 <= i <= ly - myTilde) && (mxTilde + 1 <= j <= lx - mxTilde),
          If[y[[i, j]] > 0.00, y[[i, j]], 0.00], 0.00], {i, ly}, {j, lx}],
        (* error message *)
        Message[BDRestoration::InvalidSupport];
        (* if an error occurrs, the empty matrix is returned *)
        {{}}
      ]
    ]
  ];

(*
  This function convolves two vector after having
  extended (i.e. padded with zeros) them.
*)
convolveVectors =
  Compile[{{f, _Complex, 1}, {h, _Complex, 1}},
    Module[{l, res},
      l = Length[f] + Length[h] - 1;
      (* convolve the vectors taking advantage of the convolution theorem *)
      Sqrt[l] * InverseFourier[Fourier[PadRight[f, l]] * Fourier[PadRight[h, l]]]
    ], {{Fourier[_], _Complex, 1}, {InverseFourier[_], _Complex, 1}}
  ];

(*
  This function convolves two matrices after having
  extended (i.e. padded with zeros) them.
*)
convolveMatrices =
  Compile[{{f, _Complex, 2}, {h, _Complex, 2}},
    Module[{lx, ly},
      {ly, lx} = Dimensions[f] + Dimensions[h] - 1;
      (* convolve the two matrices taking advantage of the convolution theorem *)
      Sqrt[ly * lx] * InverseFourier[Transpose[Fourier[Map[PadRight[#, ly] &,
        Transpose[Map[PadRight[#, lx] &, f]]]] * Fourier[Map[PadRight[#, ly] &,
        Transpose[Map[PadRight[#, lx] &, h]]]]]]
    ], {{Fourier[_], _Complex, 2}, {InverseFourier[_], _Complex, 2}}
  ];

(*
  This function computes the value of the cost function.
  Its parameters are: the so-far-estimated inverse filter,
  the observed and noisy image, the parameter vector and
  the estimated support.
*)
jRegValue =
  Compile[{{s, _Real, 2}, {g, _Real, 2}, {param, _Real, 1}, {nyTilde, _Integer}, {nxTilde, _Integer}}, 
    Module[{ly, lx, myHat, mxHat, nyHat, nxHat, sExt, hEst, toTrunc, tmpVal,
      tmpVal2, myTilde, mxTilde, alpha1, alpha2, alpha3, alpha4, y},
      (* get the dimensions of the input matrix *)
      {ly, lx} = Dimensions[g];
      {myHat, mxHat} = Floor[(Dimensions[s] - 1) / 2];
      {nyHat, nxHat} = {ly, lx} - 2 * {myHat, mxHat};
      {myTilde, mxTilde} = Floor[({ly, lx} - {nyTilde, nxTilde}) / 2];
      (* extend the inverse filter matrix *)
      sExt = extendPSFMatrix[s, nyHat, nxHat];
      (* compute the spectrum of the inverse filter *)
      toTrunc = Fourier[sExt];
      (* compute the estimated non-costrained restored image *)
      y = Sqrt[ly * lx] * Re[InverseFourier[Fourier[g] * toTrunc]];
      (* apply the support and non-negativity constraints on the estimated image *)
      tmpVal = y - supportConstraints[y, nyTilde, nxTilde];
      (* compute non-negativity and support constraints penalty *)
      alpha1 = Apply[Plus, tmpVal^2, {0, 1}];
      If[param[[2]] != 0,
        (* apply the spectrum truncation *)
        tmpVal2 = toTrunc - truncateComponents[toTrunc, param[[1]]];
        (* compute the truncation penalty *)
        alpha2 = param[[2]] * Apply[Plus, Map[Abs[#]^2 &, tmpVal2], {0, 1}],
        alpha2 = 0.00;
      ];
      (* compute the energy conservation penalty *)
      alpha3 = param[[3]] * (Apply[Plus, s, {0, 1}] - 1.00)^2;
      If[param[[4]] != 0,
        hEst = Take[InverseFourier[1.00 / toTrunc], {myTilde + 2,
          myTilde + nyTilde}, {mxTilde + 2, mxTilde + nxTilde}] / (ly * lx);
        (* compute the support constraint penalty on the PSF *)
        alpha4 = param[[4]] * Apply[Plus, hEst^2, {0, 1}],
        alpha4 = 0.00;
      ];
      (* due small approximation error this number might be complex. We
      know it should always be real so we can discard the immaginary part *)
      Re[alpha1 + alpha2 + alpha3 + alpha4]
    ], {{Fourier[_], _Complex, 2}, {InverseFourier[_], _Complex, 2}}
  ];

(*
  This function computes the value of the cost function and
  its gradient. Its parameters are: the so-far-estimated
  inverse filter, the observed and noisy image, the parameter
  vector and the estimated support. This function shares a
  great deal of computation with ``jRegValue'' so if we need both
  the value of the cost function and its gradient, with this
  function, we can save a lot of computation (i.e. CPU time).
*)
jRegValueGradient =
  Compile[{{s, _Real, 2}, {g, _Real, 2}, {param, _Real, 1}, {nyTilde, _Integer}, {nxTilde, _Integer}},
    Module[{ly, lx, myHat, mxHat, nyHat, nxHat, sExt, hEst, toTrunc, tmpVal, tmpVal2,
      myTilde, mxTilde, alpha1, alpha2, alpha3, alpha4, grad1, grad2, grad3, grad4, tmpGrad3, y},
      (* get the dimensions of the input matrix *)
      {ly, lx} = Dimensions[g];
      {myHat, mxHat} = Floor[(Dimensions[s] - 1) / 2];
      {nyHat, nxHat} = {ly, lx} - 2 * {myHat, mxHat};
      {myTilde, mxTilde} = Floor[({ly, lx} - {nyTilde, nxTilde}) / 2];
      (* extend the inverse filter matrix *)
      sExt = extendPSFMatrix[s, nyHat, nxHat];
      (* compute the spectrum of the inverse filter *)
      toTrunc = Fourier[sExt];
      (* compute the estimated non-costrained restored image *)
      y = Sqrt[ly * lx] * Re[InverseFourier[Fourier[g] * toTrunc]];
      (* apply the support and non-negativity constraints on the estimated image *)
      tmpVal = y - supportConstraints[y, nyTilde, nxTilde];
      (* compute non-negativity and support constraints penalty *)
      alpha1 = Apply[Plus, tmpVal^2, {0, 1}];
      grad1 = 2 * Sqrt[ly * lx] * Take[InverseFourier[InverseFourier[RotateLeft[Map[RotateLeft[#,
        mxHat] &, g], myHat]] * Fourier[tmpVal]], (2 * myHat + 1), (2 * mxHat + 1)];
      If[param[[2]] != 0,
        (* apply the spectrum truncation *)
        tmpVal2 = toTrunc - truncateComponents[toTrunc, param[[1]]];
        (* compute the truncation penalty *)
        alpha2 = param[[2]] * Apply[Plus, Map[Abs[#]^2 &, tmpVal2], {0, 1}];
        (* compute the spectrum truncation penalty gradient *)
        grad2 = 2 * param[[2]] * Re[Take[RotateRight[Map[RotateRight[#, mxHat] &,
          InverseFourier[tmpVal2]], myHat], 2 * myHat + 1, 2 * mxHat + 1]],
        alpha2 = 0.00;
        grad2 = Table[0.00, {2 * myHat + 1}, {2 * mxHat + 1}];
      ];
      (* compute the energy conservation penalty *)
      alpha3 = param[[3]] * (Apply[Plus, s, {0, 1}] - 1.00)^2;
      (* we compute this once for all and store it in a temporary variable *)
      tmpGrad3 = 2 * param[[3]] * (Apply[Plus, s, {0, 1}] - 1.00);
      (* build the energy conservation penalty gradient *)
      grad3 = Table[tmpGrad3, {2 * myHat + 1}, {2 * mxHat + 1}];
      If[param[[4]] != 0,
        hEst = Take[InverseFourier[1.00 / toTrunc], {myTilde + 2,
          myTilde + nyTilde}, {mxTilde + 2, mxTilde + nxTilde}] / (ly * lx);
        (* compute the support constraint penalty on the PSF *)
        alpha4 = param[[4]] * Apply[Plus, hEst^2, {0, 1}];
        (* compute the support constraint penalty gradient on the PSF *)
        hEst = Join[Table[0, {myTilde + 1}, {lx}], Map[PadRight[#, lx] &,
          Map[PadLeft[#, mxTilde + nxTilde] &, hEst]], Table[0, {myTilde}, {lx}]];
        grad4 = -2 * param[[4]] * Take[RotateRight[Map[RotateRight[#, mxHat] &,
          Fourier[InverseFourier[hEst] / toTrunc^2]], myHat], 2 * myHat + 1, 2 * mxHat + 1] / (ly * lx),
        alpha4 = 0.00;
        grad4 = Table[0.00, {2 * myHat + 1}, {2 * mxHat + 1}];
      ];
      (* due small approximation error this number might be complex. We
      know it should always be real so we can discard the immaginary part *)
      Re[
        (* since a compiled function cannot return a tenson of objects of
        unequal size, we cannot return a scalar and a matrix. To solve this
        problem we add a row to the gradient matrix; the elements of this
        new row are all equal to the value of the cost function. Bear this
        in mind when this function is used!! *)
        Join[{Table[alpha1 + alpha2 + alpha3 + alpha4, {Dimensions[grad1][[2]]}]},
          grad1 + grad2 + grad3 + grad4]
      ]
    ], {{Fourier[_], _Complex, 2}, {InverseFourier[_], _Complex, 2}}
  ];

(*
  This function computes the signal-to-noise ratio between
  two matrices. The matrix s represents the signal while n
  the noise affecting it. The SNR value returned is expressed
  in Decibel (dB).
*)

computeSNR[s_?MatrixQ, n_?MatrixQ] :=
  Chop[10 * Log[10, Apply[Plus, s^2, {0, 1}] /
    Apply[Plus, n^2, {0, 1}]]];

(*
  This function is used to evaluate the support of the
  image to restore. The image is expressed by means of
  the matrix g while perc is a percent threshold that
  tells the maximum intensity level of the points that
  should be considered background. This function returns
  a list of two integers which represent the estimated
  image support.
*)

evaluateSupport =
  Compile[{{g, _Real, 2}, {percentage, _Real}},
    Module[{threshold, wy, wx, gmin, gmax, dimY, dimX, maxWY, maxWX},
      {dimY, dimX} = Dimensions[g];
      {maxWY, maxWX} = Floor[{dimY, dimX} / 2];
      (* compute the graylevel threshold *)
      {gmin, gmax} = {Min[g], Max[g]};
      threshold = (percentage / 100) * (gmax - gmin) + gmin;
      (* find the maximum width of two vertical stripes on
      the right and left area of the image where the pixel
      intensity levels are all smaller that the threshold *)
      wx = 1;
      While[(Max[{Max[Take[g, {1, dimY}, {1, wx}]], Max[Take[g, {1, dimY}, {dimX - wx + 1, dimX}]]}] < threshold) && (wx < maxWX),
        wx++;
      ];
      (* find the maximum width of two horizontal stripes on
      the top and bottom area of the image where the pixel
      intensity levels are all smaller that the threshold *)
      wy = 1;
      While[(Max[{Max[Take[g, {1, wy}, {1, dimX}]], Max[Take[g, {dimY - wy + 1, dimY}, {1, dimX}]]}] < threshold) && (wy < maxWY),
        wy++;
      ];
      (* get the estimated support from the stripe widths *)
      {dimY, dimX} - 2 * {wy, wx}
    ]
  ]

(*
  This function implements the NSH-RIF algorithm.
  Its parameters are: the observed noisy image, the
  inverse filter matrix estimated, the parameter
  vector and the estimated support. Its options set
  includes: the maximum number of iteration, the
  tolerances, the true image (used for testing
  purposes), the low-pass filter parameters and
  the flag indicating if the list of restored
  images should be returned or rather only the
  last restored image. It returns either a list
  of restored images (the images restored at each
  iteration) or only the last restored image
  and some lists of log information.
*)

Options[NSHRIF] = {NSHRIFMaxIterations -> 100, NSHRIFEpsilon -> 10^-6,
  NSHRIFTrueF -> {}, NSHRIFLowPassFilter -> {},
  NSHRIFStopEpsilon -> 10^-6, NSHRIFRestoredList -> False};

NSHRIF[g_?MatrixQ, sStart_?MatrixQ, param_List, nyTilde_Integer, nxTilde_Integer, opts___?OptionQ] := 
  Module[{maxIter, epsilon, trueF, lowPassFilterParam, stopEpsilon, checkValidity, variationValueList, errorList,
    valueList, restoredList, oldValue, gTemp, sEst, ly, lx, myHat, mxHat, iteration, newGradient, newValue,
    restored, scale, err, newDirection, beta, oldGradient, alpha, restoredListEnabled},
    (* set the option variables *)
    {maxIter, epsilon, trueF, lowPassFilterParam, stopEpsilon, restoredListEnabled} =
      {NSHRIFMaxIterations, NSHRIFEpsilon, NSHRIFTrueF, NSHRIFLowPassFilter,
      NSHRIFStopEpsilon, NSHRIFRestoredList} /. {opts} /. Options[NSHRIF];
    (* the following function can be used to check whether
    the input that the user provided is valid or not *)
    checkValidity = Function[{gObserved, s, paramJ, ny, nx, paramLP},
      Module[{result, ls1, ls2, lg1, lg2},
        result = True;
        {lsx, lsy} = Dimensions[s];
        {lgx, lgy} = Dimensions[gObserved];
        (* the observed image should not contain negative elements *)
        If[Min[gObserved] < 0.00,
          Message[BDRestoration::InvalidImage];
          result = False;
        ];
        (* the number of parameters must be four *)
        If[Length[paramJ] != 4,
          Message[BDRestoration::InvalidParamNumber];
          result = False;
        ];
        (* each parameter must be greater than or equal to zero *)
        If[Min[paramJ] < 0.00,
          Message[BDRestoration::InvalidParameters, paramJ];
          result = False;
        ];
        (* the dimensions of the estimated inverse filter must me odd *)
        If[Apply[Or, EvenQ[{lsx, lsy}]],
          result = False;
          Message[BDRestoration::InvalidInverseFilterDimensions];
        ];
        (* the observed image should not be smaller than the inverse filter *)
        If[Apply[Or, Apply[Greater, {{lsx, lsy}, {lgx, lgy}}, {1}]],
          result = False;
          Message[BDRestoration::InverseFilterTooBig];
        ];
        (* the estimated support should be smaller than the observed image *)
        If[Apply[Or, Apply[Greater, {{ny, nx}, {lgx, lgy}}, {1}]],
          result = False;
          Message[BDRestoration::SupportTooBig];
        ];
        (* if pre-filtering of the observed images is
        used, the filter parameters should be three *)
        If[Position[{0, 3}, Length[paramLP]] == {},
          result = False;
          Message[BDRestoration::InvalidParamNumber];
        ];
        result
      ]
    ];
    (* check whether the input is valid *)
    If[checkValidity[g, sStart, param, nyTilde, nxTilde, lowPassFilterParam],
      (* initialize the log lists *)
      variationValueList = errorList = valueList = {};
      If[restoredListEnabled,
      	restoredList = {};
      ];
      oldValue = 1;
      (* if filtering is required, get the filtered observed image *)
      gTemp = If[lowPassFilterParam  == {},
        g,
        lowPassMatrixFilter[g, lowPassFilterParam[[1]], lowPassFilterParam[[2]], lowPassFilterParam[[3]]]
      ];
      sEst = sStart;
      (* get the dimensions of the observed image *)
      {ly, lx} = Dimensions[gTemp];
      {myHat, mxHat} = Floor[(Dimensions[sEst] - 1) / 2];
      (* variable for the number of iterations *)
      iteration = 1;
      (* the minimization routine starts here! *)
      While[
        (* pay attention: we are doing some computation
        just before evaluationg the While-condition *)
        newGradient = jRegValueGradient[sEst, gTemp, param, nyTilde, nxTilde];
        (* here we use the trick that allows to return from a
        compiled function both a scalar and a matrix. Read the
        comment at the end of the function ``jRegValueGradient''. *)
        newValue = newGradient[[1, 1]];
        newGradient = Rest[newGradient];
        (* is the job over? NOTE: Here is the While-condition *)
        (newValue >= epsilon) && (iteration <= maxIter) && ((iteration == 1) || (((oldValue - newValue) / oldValue) >= stopEpsilon)),
        (* update the log lists and variables *)
        AppendTo[valueList, newValue];
        AppendTo[variationValueList, (oldValue - newValue) / oldValue];
        oldValue = newValue;
        (* do we have to log the relative error too? *)
        If[trueF != {},
          (* compute the so-far-restored image *)
          restored = Take[supportConstraints[Re[InverseFourier[Fourier[gTemp] * Fourier[If[Dimensions[gTemp] == Dimensions[sEst],
            RotateLeft[Map[RotateLeft[#, mxHat] &, sEst], myHat], extendPSFMatrix[sEst, ly - 2 * myHat, lx - 2 * mxHat]]]]] * Sqrt[ly * lx],
            nyTilde, nxTilde], {(ly - nyTilde) / 2 + 1, (ly - nyTilde) / 2 + nyTilde}, {(lx - nxTilde) / 2 + 1, (lx - nxTilde) / 2 + nxTilde}];
          (* compute the relative error *)
          scale = Apply[Plus, restored * trueF, {0, 1}] / Apply[Plus, restored^2, {0, 1}];
          err = Apply[Plus, (scale * restored - trueF)^2, {0, 1}] / Apply[Plus, trueF^2, {0, 1}];
          (* log the relative percentage error *)
          AppendTo[errorList, err];
        ];
        newDirection = If[iteration == 1,
          -newGradient,
          beta = Apply[Plus, newGradient * (newGradient - oldGradient), {0, 1}] / Apply[Plus, oldGradient^2, {0, 1}];
          beta * oldDirection - newGradient
        ];
        (* line minimization: we can call ``FindMinimum''
        because we know at compile time the dimension of
        the space in which to search the minimum *)
        alpha = FindMinimum[jRegValue[sEst + alpha * newDirection, gTemp, param, nyTilde, nxTilde], {alpha, {0.00, 0.10}}][[2, 1, 2]];
        (* update the minimum (i.e. the inverse filter) *)
        sEst += alpha * newDirection;
        (* increment the number of iterations done *)
        ++iteration;
        (* set the variables for the next iteration *)
        oldGradient = newGradient;
        oldDirection = newDirection;
        If[restoredListEnabled,
          (* compute the so-far-restored image and append it to a list *)
          restored = supportConstraints[Re[InverseFourier[Fourier[gTemp] * Fourier[If[Dimensions[gTemp] == Dimensions[sEst],
            RotateLeft[Map[RotateLeft[#, mxHat] &, sEst], myHat], extendPSFMatrix[sEst,
            ly - 2 * myHat, lx - 2 * mxHat]]]]] * Sqrt[ly * lx], nyTilde, nxTilde];
          (* force the overall energy of the restored
          image to be the same as the observed image *)
          scale = Apply[Plus, g, {0, 1}] / Apply[Plus, restored, {0, 1}];
          AppendTo[restoredList, scale * restored];
        ];
      ];
      If[!restoredListEnabled,
        (* compute the image at the end of the algorithm *)
        restored = supportConstraints[Re[InverseFourier[Fourier[gTemp] * Fourier[If[Dimensions[gTemp] == Dimensions[sEst],
          RotateLeft[Map[RotateLeft[#, mxHat] &, sEst], myHat], extendPSFMatrix[sEst,
          ly - 2 * myHat, lx - 2 * mxHat]]]]] * Sqrt[ly * lx], nyTilde, nxTilde];
        (* force the overall energy of the restored
        image to be the same as the observed image *)
        scale = Apply[Plus, g, {0, 1}] / Apply[Plus, restored, {0, 1}];
        restored *= scale;
      ];
      (* return the restored image or list of restored images and log information *)
      If[trueF != {},
        If[restoredListEnabled,
          {restoredList, valueList, Rest[variationValueList], errorList},
          {restored, valueList, Rest[variationValueList], errorList}
        ],
        If[restoredListEnabled,
          {restoredList, valueList, Rest[variationValueList]},
          {restored, valueList, Rest[variationValueList]}
        ]
      ],
      (* if the input is not valid, empty lists are returned *)
      If[trueF != {},
        {{}, {}, {}, {}},
        {{}, {}, {}}
      ]
    ]
  ];

(*
  This function implements the total variation (TV) regularization
  of an image. It can be used to sharpen a restored image. Its
  inputs are the image to regularize, the TV-regularization
  parameters and the options. The option ``TotalVariationTrueF''
  allows us to log the relative error between the true image and the
  TV-regularized one at each iteration of the total variation
  regularization algorithm. This function returns the regularized
  image and the list of relative errors if ``TotalVariationTrueF''
  is not set to the empty list.
*)

Options[totalVariation] = {TotalVariationTrueF -> {}, TotalVariationEpsilon -> 10^-8};

totalVariation[y_?MatrixQ, nu_?ListQ, opts___?OptionQ] :=
  Module[{f, r, gradF, newDir, oldDir, newValue, oldValue, alpha, newGrad, oldGrad, ny, nx, iter, trueF, errorList, scale, err, epsilon},
    (* get the options *)
    {trueF, epsilon} = {TotalVariationTrueF, TotalVariationEpsilon} /. {opts} /. Options[totalVariation];
    (* the length of the regularization parameters vector should be 2 *)
    If[Length[nu] == 2,
      If[Min[nu] >= 0,
        (* initialize the log list *)
        errorList = {};
        r = y;
        (* get the dimension of the image to regularize *)
        {ny, nx} = Dimensions[y];
        (* this is the function to minimize. The
        minimum is the regularized image *)
        f = Compile[{{s, _Real, 2}},
          Module[{i, j}, 
            Apply[Plus, (s - y)^2, {0, 1}] / 2.00 + nu[[1]] * Apply[Plus, Table[Sqrt[(s[[i, j]] - s[[i + 1,
              j + 1]])^2 + (s[[i + 1, j]] - s[[i , j + 1]])^2 + nu[[2]]], {i, ny - 1}, {j, nx - 1}], {0, 1}]
          ], {{nu, _Real, 1}, {y, _Real, 2}}];
        (* this is the gradient of the function to minimize *)
        gradF = Compile[{{s, _Real, 2}},
          Module[{i, j, nu1, nu2},
            {nu1, nu2} = nu;
            (s - y) + nu1 * Table[
              (* Since we use the ``oblique cross approach'' there are
              nine cases that must be handled; the reason can be seen
              selecting the first column, the middle columns and the
              last column of the image matrix and then selecting the
              top element, the middle elements and the bottom element
              of such vectors/submatrices. We will obtain nine ``blocks''
              (four ``1-by-1'' matrices, two ``1-by-(n - 2)'' matrices,
              two ``(n - 2)-by-1'' matrices and a ``(n - 2)-by-(n - 2)''
              matrix). For each block there are different elements on
              wich the gradient depends; that yields all the cases we
              should consider *)
              Which[
                i == 1,
                Which[
                  j == 1,
                  (s[[i, j]] - s[[i + 1, j + 1]]) / Sqrt[nu2 + (s[[i + 1, j]] - 
                    s[[i, j + 1]])^2 + (s[[i, j]] - s[[i + 1, j + 1]])^2],
                  j == nx,
                  -(s[[i + 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i + 1, j - 1]] - 
                    s[[i, j]])^2 + (s[[i, j - 1]] - s[[i + 1, j]])^2],
                  True,
                  (s[[i, j]] - s[[i + 1, j + 1]]) / Sqrt[nu2 + (s[[i + 1, j]] - 
                    s[[i, j + 1]])^2 + (s[[i, j]] - s[[i + 1, j + 1]])^2] -
                    (s[[i + 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i + 1, j - 1]] - 
                    s[[i, j]])^2 + (s[[i, j - 1]] - s[[i + 1, j]])^2]
                ],
                i == ny,
                Which[
                  j == 1,
                  (s[[i, j]] - s[[i - 1, j + 1]]) / Sqrt[nu2 + (s[[i, j]] - 
                    s[[i - 1, j + 1]])^2 + (s[[i - 1, j]] - s[[i, j + 1]])^2],
                  j == nx,
                  -(s[[i - 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i, j - 1]] - 
                    s[[i - 1, j]])^2 + (s[[i - 1, j - 1]] - s[[i, j]])^2],
                  True,
                  (s[[i, j]] - s[[i - 1, j + 1]]) / Sqrt[nu2 + (s[[i, j]] - 
                    s[[i - 1, j + 1]])^2 + (s[[i - 1, j]] - s[[i, j + 1]])^2] -
                    (s[[i - 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i, j - 1]] - 
                    s[[i - 1, j]])^2 + (s[[i - 1, j - 1]] - s[[i, j]])^2]
                ],
                True,
                Which[
                  j == 1,
                  (s[[i, j]] - s[[i - 1, j + 1]]) / Sqrt[nu2 + (s[[i, j]] - 
                    s[[i - 1, j + 1]])^2 + (s[[i - 1, j]] - s[[i, j + 1]])^2] +
                    (s[[i, j]] - s[[i + 1, j + 1]]) / Sqrt[nu2 + (s[[i + 1, j]] - 
                    s[[i, j + 1]])^2 + (s[[i, j]] - s[[i + 1, j + 1]])^2],
                  j == nx,
                  -(s[[i - 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i, j - 1]] - 
                    s[[i - 1, j]])^2 + (s[[i - 1, j - 1]] - s[[i, j]])^2] -
                    (s[[i + 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i + 1, j - 1]] - 
                    s[[i, j]])^2 + (s[[i, j - 1]] - s[[i + 1, j]])^2],
                  True,
                  -(s[[i - 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i, j - 1]] - 
                    s[[i - 1, j]])^2 + (s[[i - 1, j - 1]] - s[[i, j]])^2] +
                    (s[[i, j]] - s[[i - 1, j + 1]]) / Sqrt[nu2 + (s[[i, j]] - 
                    s[[i - 1, j + 1]])^2 + (s[[i - 1, j]] - s[[i, j + 1]])^2] -
                    (s[[i + 1, j - 1]] - s[[i, j]]) / Sqrt[nu2 + (s[[i + 1, j - 1]] - 
                    s[[i, j]])^2 + (s[[i, j - 1]] - s[[i + 1, j]])^2] +
                    (s[[i, j]] - s[[i + 1, j + 1]]) / Sqrt[nu2 + (s[[i + 1, j]] - 
                    s[[i, j + 1]])^2 + (s[[i, j]] - s[[i + 1, j + 1]])^2]
                ]
              ], {i, ny}, {j, nx}
            ]
          ], {{nu, _Real, 1}, {y, _Real, 2}}];
        (* get the gradient *)
        newGrad = gradF[r];
        (* is the image already OK? *)
        If[Apply[Plus, newGrad^2, {0, 1}] > epsilon,
          newValue = f[r];
          (* compute the minimization direction *)
          newDir = -newGrad;
          (* compute the first line minimization *)
          alpha = FindMinimum[f[r + alpha * newDir], {alpha, {0.00, 0.20}}, MaxIterations -> 2000,
            AccuracyGoal -> 16, PrecisionGoal -> 16, WorkingPrecision -> 32][[2, 1, 2]];
          (* update the regularized image *)
          r += alpha * newDir;
          iter = 1;
          While[
            (* some computation before evaluating the While-condition *)
            oldDir = newDir;
            oldValue = newValue;
            oldGrad = newGrad;
            newValue = f[r];
            newGrad = gradF[r];
            (* is our work over? NOTE: here is the While-condition *)
            (Abs[oldValue - newValue] > epsilon) && (Apply[Plus, newGrad^2, {0, 1}] > epsilon) && (iter < 1000),
            oldDir = newDir;
            (* compute the minimization direction *)
            newDir = (Apply[Plus, newGrad * (newGrad - oldGrad), {0, 1}] / Apply[Plus, oldGrad^2, {0, 1}]) * oldDir - newGrad;
            (* compute a line minimization *)
            alpha = FindMinimum[f[r + alpha * newDir], {alpha, {0.00, 0.20}}, MaxIterations -> 2000,
              AccuracyGoal -> 16, PrecisionGoal -> 16, WorkingPrecision -> 32][[2, 1, 2]];
            (* check if log information must be computed *)
            If[trueF != {},
              (* compute the relative error *)
              scale = Apply[Plus, r * trueF, {0, 1}] / Apply[Plus, r^2, {0, 1}];
              err = Apply[Plus, (scale * r - trueF)^2, {0, 1}] / Apply[Plus, trueF^2, {0, 1}];
              (* append the computed percentage error to the log list *)
              AppendTo[errorList, err];
            ];
            (* update the regularized image *)
            r += alpha * newDir;
            (* increment the number of iteration done *)
            iter++;
          ];
        ];
        If[trueF != {},
          (* return the regularized image and the log list *)
          {r, errorList},
          (* return the regularized image only *)
          r
        ],
        Message[BDRestoration::InvalidParameters, nu];
      ],
      Message[BDRestoration::InvalidParamNumber];
      (* the empty matrix is returned if the parameters vector is not valid *)
      {{}}
    ]
  ];

End[]

EndPackage[]
